import re
import asyncio
import logging
from enum import Enum
from datetime import datetime, timezone
from bson.codec_options import CodecOptions
from pymongo import DESCENDING, ASCENDING


log = logging.getLogger(__name__)


class Ordering(Enum):

    title_asc = ('template.title', ASCENDING)
    title_desc = ('template.title', DESCENDING)
    start_asc = ('start', ASCENDING)
    start_desc = ('start', DESCENDING)
    end_asc = ('end', ASCENDING)
    end_desc = ('end', DESCENDING)

    @classmethod
    def keys(cls):
        return [key for key in cls.__members__.keys()]


class WorkflowInstancesCollection:

    REQUESTER_REGEX = re.compile(r'^nyuki://.*')

    def __init__(self, db):
        # Handle timezones in mongo collections.
        # See http://api.mongodb.com/python/current/examples/datetimes.html#reading-time
        self._instances = db['workflow_instances'].with_options(
            codec_options=CodecOptions(tz_aware=True, tzinfo=timezone.utc)
        )

    async def index(self):
        # Workflow
        await self._instances.create_index('id', unique=True)
        await self._instances.create_index('state')
        await self._instances.create_index('requester')
        # Search and sorting indexes
        await self._instances.create_index('template.title')
        await self._instances.create_index([('start', DESCENDING)])
        await self._instances.create_index([('end', DESCENDING)])

    async def get_one(self, instance_id, full=False):
        """
        Return the instance with `instance_id` from workflow history.
        """
        return await self._instances.find_one({'id': instance_id}, {'_id': 0})

    async def get(self, root=False, full=False, offset=None, limit=None,
                  since=None, state=None, search=None, order=None):
        """
        Return all instances from history from `since` with state `state`.
        """
        query = {}
        # Prepare query
        if isinstance(since, datetime):
            query['start'] = {'$gte': since}
        if isinstance(state, Enum):
            query['state'] = state.value
        if root is True:
            query['requester'] = {'$not': self.REQUESTER_REGEX}
        if search:
            query['template.title'] = {'$regex': '.*{}.*'.format(search)}

        cursor = self._instances.find(query, {'_id': 0})
        # Count total results regardless of limit/offset
        count = await cursor.count()

        # Sort depending on Order enum values
        if order is not None:
            cursor.sort(*order)
        else:
            # End descending by default
            cursor.sort(*Ordering.end_desc.value)

        # Set offset and limit
        if isinstance(offset, int) and offset >= 0:
            cursor.skip(offset)
        if isinstance(limit, int) and limit > 0:
            cursor.limit(limit)

        # Execute query
        return count, await cursor.to_list(None)

    async def insert(self, workflow):
        """
        Insert a finished workflow report into the workflow history.
        """
        await self._instances.insert_one(workflow)



import asyncio
import logging


log = logging.getLogger(__name__)


class TriggerCollection:

    def __init__(self, db):
        self._triggers = db['triggers']

    async def index(self):
        await self._triggers.create_index('tid', unique=True)

    async def get(self):
        """
        Return a list of all trigger forms
        """
        cursor = self._triggers.find(None, {'_id': 0})
        return await cursor.to_list(None)

    async def get_one(self, template_id):
        """
        Return the trigger form of a given workflow template id
        """
        return await self._triggers.find_one({'tid': template_id}, {'_id': 0})

    async def insert(self, tid, form):
        """
        Insert a trigger form for the given workflow template
        """
        data = {'tid': tid, 'form': form}
        await self._triggers.replace_one({'tid': tid}, data, upsert=True)
        return data

    async def delete(self, tid):
        """
        Delete a trigger form
        """
        await self._triggers.delete_one({'tid': tid})



import asyncio
import logging
from pymongo import ASCENDING, DESCENDING


log = logging.getLogger(__name__)


class TaskTemplatesCollection:

    """
    {
        "id": <str>,
        "name": <str>,
        "config": {},
        "topics": [<str>],
        "title": <str>,
        "workflow_template": {
            "id": <uuid4>,
            "version": <int>
        }
    }
    """

    def __init__(self, db):
        self._templates = db['task_templates']

    async def index(self):
        # Pair of indexes on the workflow template id/version
        await self._templates.create_index([
            ('id', ASCENDING),
            ('workflow_template.id', ASCENDING),
            ('workflow_template.version', DESCENDING),
        ], unique=True)

    async def get(self, workflow_id, version):
        """
        Return all the task template of a workflow and a version.
        """
        cursor = self._templates.find(
            {
                'workflow_template.id': workflow_id,
                'workflow_template.version': version,
            },
            {
                '_id': 0,
                'workflow_template': 0,
            },
        )
        return await cursor.to_list(None)

    async def insert_many(self, tasks, template):
        """
        Update multiple tasks at once, remove the now unused tasks.
        """
        # Delete all tasks that are not in the list (draft mode).
        await self._templates.delete_many({
            'id': {'$nin': [task['id'] for task in tasks]},
            'workflow_template.id': template['id'],
            'workflow_template.version': template['version'],
        })

        # Re-update all the tasks.
        must_execute = False
        bulk = self._templates.initialize_unordered_bulk_op()
        for task in tasks:
            must_execute = True
            task['workflow_template'] = {
                'id': template['id'],
                'version': template['version'],
            }
            bulk.find({
                'id': task['id'],
                'workflow_template.id': template['id'],
                'workflow_template.version': template['version'],
            }).upsert().replace_one(task)

        if must_execute is True:
            await bulk.execute()

    async def delete_many(self, tid):
        """
        Delete all tasks of one template.
        """
        await self._templates.delete_many({'workflow_template.id': tid})



import logging
from pymongo.errors import AutoReconnect, DuplicateKeyError
from tukio.workflow import TemplateGraphError, WorkflowTemplate

from nyuki.api import Response, resource
from nyuki.workflow.validation import validate, TemplateError
from nyuki.workflow.db.workflow_templates import TemplateState


log = logging.getLogger(__name__)


@resource('/workflow/tasks', versions=['v1'])
class ApiTasks:

    async def get(self, request):
        """
        Return the available tasks
        """
        return Response(self.nyuki.AVAILABLE_TASKS)


class _TemplateResource:

    """
    Share methods between templates resources
    """

    async def upsert_draft(self, template, request):
        """
        Helper to insert/update a draft.
        """
        template = {
            **template.as_dict(),
            'title': request['title'],
            'tags': request.get('tags', []),
        }

        # Store task extra info (ie. title)
        rqst_tasks = request.get('tasks', [])
        tmpl_tasks = template['tasks']
        for src in rqst_tasks:
            match = list(filter(lambda t: t['id'] == src['id'], tmpl_tasks))
            if match:
                match[0].update({'title': src.get('title')})

        try:
            return await self.nyuki.storage.upsert_draft(template)
        except DuplicateKeyError as exc:
            raise DuplicateKeyError('Template already exists for this version') from exc

    def errors_from_validation(self, template):
        """
        Validate and return the list of errors if any
        """
        errors = None
        try:
            validate(template)
        except TemplateError as err:
            errors = err.as_dict()
        return errors


@resource('/workflow/templates', versions=['v1'])
class ApiTemplates(_TemplateResource):

    async def get(self, request):
        """
        Return available workflows' DAGs
        """
        try:
            templates = await self.nyuki.storage.get_templates(
                full=(request.query.get('full') == '1'),
            )
        except AutoReconnect:
            return Response(status=503)
        return Response(templates)

    async def put(self, request):
        """
        Create a workflow DAG from JSON
        """
        request = await request.json()
        if 'id' in request:
            del request['id']

        if 'title' not in request:
            return Response(status=400, body={
                'error': "workflow 'title' is mandatory"
            })

        if self.nyuki.DEFAULT_POLICY is not None and 'policy' not in request:
            request['policy'] = self.nyuki.DEFAULT_POLICY

        try:
            template = WorkflowTemplate.from_dict(request)
        except TemplateGraphError as exc:
            return Response(status=400, body={
                'error': str(exc)
            })

        try:
            tmpl_dict = await self.upsert_draft(template, request)
        except DuplicateKeyError as exc:
            return Response(status=409, body={
                'error': exc
            })

        tmpl_dict['errors'] = self.errors_from_validation(template)
        return Response(tmpl_dict)


@resource('/workflow/templates/{tid}', versions=['v1'])
class ApiTemplate(_TemplateResource):

    async def get(self, request, tid):
        """
        Return the latest version of the template
        """
        try:
            tmpl = await self.nyuki.storage.get_templates(tid, full=True)
        except AutoReconnect:
            return Response(status=503)
        if not tmpl:
            return Response(status=404)
        return Response(tmpl)

    async def put(self, request, tid):
        """
        Create a new draft for this template id.
        """
        try:
            tmpl = await self.nyuki.storage.get_template(tid)
        except AutoReconnect:
            return Response(status=503)
        if not tmpl:
            return Response(status=404)

        draft = await self.nyuki.storage.get_template(tid, draft=True)
        if draft:
            return Response(status=409, body={
                'error': 'This draft already exists'
            })

        request = await request.json()

        try:
            # Set template ID from url
            template = WorkflowTemplate.from_dict({**request, 'id': tid})
        except TemplateGraphError as exc:
            return Response(status=400, body={
                'error': str(exc)
            })

        try:
            tmpl_dict = await self.upsert_draft(template, request)
        except DuplicateKeyError as exc:
            return Response(status=409, body={
                'error': exc
            })

        tmpl_dict['errors'] = self.errors_from_validation(template)
        return Response(tmpl_dict)

    async def patch(self, request, tid):
        """
        Modify the template's metadata
        """
        try:
            templates = await self.nyuki.storage.get_templates(template_id=tid)
        except AutoReconnect:
            return Response(status=503)
        if not templates:
            return Response(status=404)

        request = await request.json()

        # Add ID, request dict cleaned in storage
        metadata = await self.nyuki.storage.update_workflow_metadata(tid, {
            'title': request.get('title'),
            'tags': request.get('tags', []),
        })

        return Response(metadata)

    async def delete(self, request, tid):
        """
        Delete the template and all its versions.
        """
        try:
            templates = await self.nyuki.storage.get_templates(
                template_id=tid, full=True
            )
        except AutoReconnect:
            return Response(status=503)
        if not templates:
            return Response(status=404)

        await self.nyuki.storage.delete_template(tid)
        return Response(templates)


@resource('/workflow/templates/{tid}/{version:\d+}', versions=['v1'])
class ApiTemplateVersion(_TemplateResource):

    async def get(self, request, tid, version):
        """
        Return the template's given version
        """
        try:
            tmpl = await self.nyuki.storage.get_template(tid, version=version)
        except AutoReconnect:
            return Response(status=503)
        if not tmpl:
            return Response(status=404)

        return Response(tmpl)


@resource('/workflow/templates/{tid}/draft', versions=['v1'])
class ApiTemplateDraft(_TemplateResource):

    async def get(self, request, tid):
        """
        Return the template's draft, if any
        """
        try:
            tmpl = await self.nyuki.storage.get_template(tid, draft=True)
        except AutoReconnect:
            return Response(status=503)
        if not tmpl:
            return Response(status=404)

        return Response(tmpl)

    async def post(self, request, tid):
        """
        Publish a draft into production
        """
        try:
            tmpl_dict = await self.nyuki.storage.get_template(tid, draft=True)
        except AutoReconnect:
            return Response(status=503)
        if not tmpl_dict:
            return Response(status=404)

        try:
            # Set template ID from url
            template = WorkflowTemplate.from_dict(tmpl_dict)
        except TemplateGraphError as exc:
            return Response(status=400, body={
                'error': str(exc)
            })

        errors = self.errors_from_validation(template)
        if errors is not None:
            return Response(errors, status=400)

        # Update draft into a new template
        await self.nyuki.storage.publish_draft(tid)
        tmpl_dict['state'] = TemplateState.ACTIVE.value
        return Response(tmpl_dict)

    async def patch(self, request, tid):
        """
        Modify the template's draft
        """
        try:
            tmpl = await self.nyuki.storage.get_template(tid, draft=True)
        except AutoReconnect:
            return Response(status=503)
        if not tmpl:
            return Response(status=404)

        request = await request.json()

        try:
            # Set template ID from url
            template = WorkflowTemplate.from_dict({**request, 'id': tid})
        except TemplateGraphError as exc:
            return Response(status=400, body={
                'error': str(exc)
            })

        try:
            tmpl_dict = await self.upsert_draft(template, request)
        except DuplicateKeyError as exc:
            return Response(status=409, body={
                'error': str(exc)
            })

        tmpl_dict['errors'] = self.errors_from_validation(template)
        return Response(tmpl_dict)

    async def delete(self, request, tid):
        """
        Delete the template's draft
        """
        try:
            tmpl = await self.nyuki.storage.get_template(tid, draft=True)
        except AutoReconnect:
            return Response(status=503)
        if not tmpl:
            return Response(status=404)

        await self.nyuki.storage.delete_template(tid, draft=True)
        return Response(tmpl)



import logging


log = logging.getLogger(__name__)


class DataProcessingCollection:

    def __init__(self, db, collection_name):
        self._rules = db[collection_name]

    async def index(self):
        await self._rules.create_index('id', unique=True)

    async def get(self):
        """
        Return a list of all rules
        """
        cursor = self._rules.find(None, {'_id': 0})
        return await cursor.to_list(None)

    async def get_one(self, rule_id):
        """
        Return the rule for given id or None
        """
        return await self._rules.find_one({'id': rule_id}, {'_id': 0})

    async def insert(self, data):
        """
        Insert a new data processing rule:
        {
            "id": "rule_id",
            "name": "rule_name",
            "config": {
                "some": "configuration"
            }
        }
        """
        query = {'id': data['id']}
        log.info(
            "Inserting data processing rule in collection '%s'",
            self._rules.name
        )
        log.debug('upserting data: %s', data)
        await self._rules.replace_one(query, data, upsert=True)

    async def delete(self, rule_id=None):
        """
        Delete a rule from its id or all rules
        """
        query = {'id': rule_id} if rule_id is not None else None
        log.info("Removing rule(s) from collection '%s'", self._rules.name)
        log.debug('delete query: %s', query)
        await self._rules.delete_one(query)



from tukio.task import TaskTemplate
from tukio.workflow import WorkflowTemplate


class WorkflowSelector:

    def __init__(self, storage):
        self.storage = storage

    async def get(self, tmpl_id):
        template = await self.storage.get_template(
            tmpl_id, draft=False
        )
        return WorkflowTemplate.from_dict(template)

    async def select(self, topic):
        templates = await self.storage.get_for_topic(topic)
        return [
            WorkflowTemplate.from_dict(template)
            for template in templates
        ]



import asyncio
import logging
from enum import Enum
from pymongo import DESCENDING


log = logging.getLogger(__name__)


class TemplateState(Enum):

    DRAFT = 'draft'
    ACTIVE = 'active'
    ARCHIVED = 'archived'

    @classmethod
    def active_states(cls):
        return [cls.DRAFT.value, cls.ACTIVE.value]

    @classmethod
    def draft_state(cls, draft):
        return cls.DRAFT.value if draft is True else cls.ACTIVE.value


class WorkflowTemplatesCollection:

    """
    Holds all the templates created for tukio, with their versions.
    These records will be used to ensure a persistence of created workflows
    in case the nyuki get into trouble.
    Templates are retrieved and loaded at startup.

    {
        "id": <uuid4>,
        "policy": <str>,
        "topics": [<str>],
        "graph": {},
        "version": <int>,
        "state": <draft | active | archived>
    }
    """

    def __init__(self, db):
        self._templates = db['workflow_templates']

    async def index(self):
        await self._templates.create_index('topics')
        await self._templates.create_index(
            [('id', DESCENDING), ('version', DESCENDING)],
            unique=True
        )
        await self._templates.create_index(
            [('id', DESCENDING), ('state', DESCENDING)]
        )

    async def get(self, template_id=None, full=False):
        """
        Return all active and draft templates
        Used at nyuki's startup and GET /v1/templates
        """
        query = {'state': {'$in': TemplateState.active_states()}}
        if template_id is not None:
            query['id'] = template_id
        filters = {'_id': 0}
        if full is False:
            filters.update({'id': 1, 'state': 1, 'version': 1, 'topics': 1})

        # Retrieve only the actives and the drafts
        cursor = self._templates.find(query, filters)
        return await cursor.to_list(None)

    async def get_one(self, tid, version=None, draft=False):
        """
        Return a template's configuration and versions
        """
        if version is not None:
            # We ask for a specific version, regardless of its state.
            query = {'id': tid, 'version': int(version)}
        else:
            # Else, we ask for either the active version or the draft.
            query = {
                'id': tid,
                'state': TemplateState.draft_state(draft),
            }

        return await self._templates.find_one(query, {'_id': 0})

    async def get_for_topic(self, topic):
        """
        Return the latest templates (non-draft) that wait
        for a certain topic.
        """
        query = {
            '$or': [
                {'topics': topic, 'state': TemplateState.ACTIVE.value},
                {'topics': None, 'state': TemplateState.ACTIVE.value},
            ]
        }
        cursor = self._templates.find(query, {'_id': 0})
        return await cursor.to_list(None)

    async def get_last_version(self, tid):
        """
        Return the highest version of a template
        """
        query = {'id': tid, 'state': TemplateState.ACTIVE.value}
        template = await self._templates.find_one(query, {'version': 1})
        return template['version'] if template else 0

    async def insert_draft(self, template):
        """
        Insert or replace a template document.
        This is used to update a draft.
        """
        query = {
            'id': template['id'],
            'state': TemplateState.DRAFT.value,
        }
        await self._templates.replace_one(query, template, upsert=True)

    async def publish_draft(self, tid):
        """
        Set the last published template to 'archived'.
        Set the draft of this template to 'active'.
        """
        await self._templates.update_one(
            {'id': tid, 'state': TemplateState.ACTIVE.value},
            {'$set': {'state': TemplateState.ARCHIVED.value}},
        )
        await self._templates.update_one(
            {'id': tid, 'state': TemplateState.DRAFT.value},
            {'$set': {'state': TemplateState.ACTIVE.value}},
        )

    async def delete(self, tid, draft=False):
        """
        Delete a template with all its versions.
        Delete only the draft if specified.
        """
        if draft is False:
            # Delete all template versions.
            await self._templates.delete_many({'id': tid})
            return

        # Delete only the draft.
        await self._templates.delete_one({
            'id': tid,
            'state': TemplateState.DRAFT.value,
        })



import logging
from datetime import timezone
from bson.codec_options import CodecOptions


log = logging.getLogger(__name__)
WS_FILTERS = ('quorum', 'status', 'twilio_error', 'diff')


class TaskInstancesCollection:

    TASK_HISTORY_FILTERS = {
        '_id': 0,
        # Tasks
        'template.id': 1,
        'template.name': 1,
        'template.config': 1,
        'template.topics': 1,
        'template.title': 1,
        'template.timeout': 1,
        # Exec
        'id': 1,
        'start': 1,
        'end': 1,
        'state': 1,
        # Graph-specific data fields
        **{'outputs.{}'.format(key): 1 for key in WS_FILTERS}
    }

    def __init__(self, db):
        # Handle timezones in mongo collections.
        # See http://api.mongodb.com/python/current/examples/datetimes.html#reading-time
        self._instances = db['task_instances'].with_options(
            codec_options=CodecOptions(tz_aware=True, tzinfo=timezone.utc)
        )

    async def index(self):
        await self._instances.create_index('id', unique=True)
        await self._instances.create_index('workflow_instance_id')

    async def get(self, wid, full=False):
        """
        Return all task instances of one workflow.
        """
        if full is False:
            filters = self.TASK_HISTORY_FILTERS
        else:
            filters = {'_id': 0, 'workflow_instance_id': 0}
        cursor = self._instances.find({'workflow_instance_id': wid}, filters)
        return await cursor.to_list(None)

    async def get_one(self, tid, full=False):
        """
        Return one task instance.
        """
        filters = {'_id': 0, 'workflow_instance_id': 0}
        if full is False:
            filters.update({'inputs': 0, 'outputs': 0})
        return await self._instances.find_one({'id': tid}, filters)

    def get_for_csv(self, workflow_id):
        """
        Return the data needed for a CSV import of one workflow execution.
        To be used with `async for task in get_for_csv():`
        """
        return self._instances.find({
            'workflow_instance_id': workflow_id,
            'template.name': {'$in': ['send_sms', 'call', 'send_email', 'wait_sms', 'wait_call', 'wait_email']},
            'reporting': {'$ne': None},
        }, {
            'start': 1,
            'template.id': 1,
            'template.title': 1,
            'template.name': 1,
            'reporting.contacts': 1,
        })

    async def get_data(self, tid):
        """
        Return the data (inputs/outputs) of one executed task.
        """
        return await self._instances.find_one(
            {'id': tid},
            {'_id': 0, 'inputs': 1, 'outputs': 1},
        )

    async def insert_many(self, tasks):
        """
        Insert all the tasks of a finished workflow.
        """
        await self._instances.insert_many(tasks)



import time
import asyncio
import logging
from math import ceil
from uuid import uuid4
from pymongo import ASCENDING, DESCENDING
from pymongo.errors import BulkWriteError
from motor.motor_asyncio import AsyncIOMotorClient


log = logging.getLogger(__name__)


def timed(func):
    async def wraps(*args, **kwargs):
        start = time.time()
        r = await func(*args, **kwargs)
        log.debug("'%s' took %s s", func.__name__, time.time() - start)
        return r
    return wraps


class Migrator:

    """
    Used to fill a collection and progressively erase an old one.
    """

    def __init__(self, old_collection, new_collection):
        self._old_collection = old_collection
        self._new_collection = new_collection
        self._old_bulk = old_collection.initialize_unordered_bulk_op()
        self._new_bulk = new_collection.initialize_unordered_bulk_op()
        self._cursor = old_collection.find()
        self._batch_size = 0
        self.count = 0

    @property
    def old(self):
        return self._old_bulk

    @property
    def new(self):
        return self._new_bulk

    def __aiter__(self):
        return self

    async def __anext__(self):
        self._batch_size += 1
        self.count += 1

        # Insert in bulk. The batch size is a cpu/memory compromise.
        if self._batch_size == 200:
            await asyncio.wait([
                asyncio.ensure_future(self._old_bulk.execute()),
                asyncio.ensure_future(self._new_bulk.execute()),
            ])
            self._batch_size = 0
            self._old_bulk = self._old_collection.initialize_unordered_bulk_op()
            self._new_bulk = self._new_collection.initialize_unordered_bulk_op()

        # Return one document.
        await self._cursor.fetch_next
        doc = self._cursor.next_object()

        # End of iteration, insert the rest.
        if not doc:
            if self._batch_size > 0:
                await asyncio.wait([
                    asyncio.ensure_future(self._old_bulk.execute()),
                    asyncio.ensure_future(self._new_bulk.execute()),
                ])
            await self._old_collection.drop()
            raise StopAsyncIteration

        return doc


class Migration:

    def __init__(self, host, database, validate_on_start=None, **kwargs):
        client = AsyncIOMotorClient(host, **kwargs)
        self.db = client[database]

    @timed
    async def run(self):
        """
        Run the migrations.
        """
        log.info("Starting migrations on database '%s'", self.db.name)
        collections = await self.db.collection_names()
        if 'metadata' in collections:
            await self._migrate_workflow_metadata()
        if 'templates' in collections:
            await self._migrate_workflow_templates()
        if 'workflow-instances' in collections:
            await self._migrate_workflow_instances()
        if 'task-instances' in collections:
            await self._migrate_task_instances()
        if 'instances' in collections:
            await self._migrate_old_instances()
        log.info("Migration on database '%s' passed", self.db.name)

    @timed
    async def _migrate_workflow_metadata(self):
        old_col = self.db['metadata']
        new_col = self.db['workflow_metadata']
        migrator = Migrator(old_col, new_col)
        async for metadata in migrator:
            metadata['workflow_template_id'] = metadata.pop('id')
            migrator.new.insert(metadata)
            migrator.old.find({'_id': metadata['_id']}).remove_one()
        log.info('%s workflow metadata migrated', migrator.count)

    @timed
    async def _migrate_workflow_templates(self):
        """
        Replace documents with new structure.
        This uses a particular search filter because of the 'draft' attribute.
        """
        count = 0
        template_state = [None, None]
        sort = [('id', ASCENDING), ('version', DESCENDING)]
        wf_col = self.db['workflow_templates']
        task_col = self.db['task_templates']

        # Sorted on version number to set the proper 'state' values.
        async for template in self.db['templates'].find(None, sort=sort):
            count += 1
            # Ensure we got only one 'draft' if any, one 'active' if any,
            # and the rest 'archived'.
            if template['id'] != template_state[0]:
                template_state = [template['id'], 'active']
            if template['draft'] is True:
                state = 'draft'
            else:
                state = template_state[1]
                if state == 'active':
                    template_state[1] = 'archived'

            template['state'] = state
            template['timeout'] = None
            del template['draft']
            # Remove tasks.
            tasks = template.pop('tasks')
            await wf_col.replace_one(
                {'_id': template['_id']}, template, upsert=True
            )

            # Migrate and insert task templates.
            workflow_template = {
                'id': template['id'],
                'version': template['version'],
            }
            task_bulk = task_col.initialize_unordered_bulk_op()
            for task in tasks:
                task['workflow_template'] = workflow_template
                task = self._migrate_one_task_template(task)
                task_bulk.find({
                    'id': task['id'],
                    'workflow_template.id': workflow_template['id'],
                    'workflow_template.version': workflow_template['version'],
                }).upsert().replace_one(task)
            if tasks:
                await task_bulk.execute()

        await self.db['templates'].drop()
        log.info('%s workflow templates splited and migrated', count)

    @timed
    async def _migrate_workflow_instances(self):
        """
        Replace workflow instance documents with new structure.
        """
        old_col = self.db['workflow-instances']
        new_col = self.db['workflow_instances']
        migrator = Migrator(old_col, new_col)
        async for workflow in migrator:
            # Convert workflow instance.
            instance = workflow.pop('exec')
            instance['_id'] = workflow.pop('_id')
            instance['template'] = workflow
            # Insert.
            migrator.new.insert(instance)
            migrator.old.find({'_id': instance['_id']}).remove_one()
        log.info('%s workflow instances migrated', migrator.count)

    def _migrate_one_task_template(self, template):
        """
        Do the task configuration migrations.
        Add the new 'timeout' field.
        """
        config = template['config']
        template['timeout'] = None

        if template['name'] == 'join':
            template['timeout'] = config.pop('timeout', None)

        elif template['name'] == 'trigger_workflow':
            template['timeout'] = config.get('timeout')
            template['config'] = {
                'blocking': config.get('await_completion', True),
                'template': {
                    'service': 'twilio' if 'twilio' in config.get('nyuki_api', '') else 'pipeline',
                    'id': config.get('template', ''),
                    'draft': config.get('draft', False),
                },
            }

        elif template['name'] in ['send_sms', 'call', 'wait_sms', 'wait_email', 'wait_call']:
            if 'blocking' in config:
                template['timeout'] = config['blocking']['timeout']
                config['blocking'] = True

        elif template['name'] == 'gather':
            if 'finishOnKey' in config:
                config['finish_on_key'] = config.pop('finishOnKey')
            if 'numDigits' in config:
                config['num_digits'] = config.pop('numDigits')
            # Remove finish_on_key if num_digits is 1
            if 'finish_on_key' in config and config.get('num_digits') == 1:
                del config['finish_on_key']

        elif template['name'] == 'record':
            if 'finishOnKey' in config:
                config['finish_on_key'] = config.pop('finishOnKey')
            if 'maxLength' in config:
                config['max_length'] = config.pop('maxLength')
            if 'playBeep' in config:
                config['play_beep'] = config.pop('playBeep')

        return template

    def _new_task(self, task, workflow_instance_id=None):
        """
        Migrate a task instance.
        """
        # If task was never executed, fill it with 'not-started'.
        instance = task.pop('exec') or {
            'id': str(uuid4()),
            'status': 'not-started',
            'start': None,
            'end': None,
            'inputs': None,
            'outputs': None,
            'reporting': None,
        }
        if '_id' in task:
            instance['_id'] = task.pop('_id')
        instance['workflow_instance_id'] = workflow_instance_id or task.pop('workflow_exec_id')
        instance['template'] = self._migrate_one_task_template(task)

        if instance['reporting'] is None:
            return instance

        # Reporting migrations
        if instance['template']['name'] == 'group_selector':
            rgroups = instance['reporting'].get('groups', {})
            if isinstance(rgroups, dict):
                new_reporting = []
                for gid, medias in rgroups.items():
                    new_reporting.append({
                        'name': '',
                        'medias': medias,
                        'uid': gid,
                    })
                instance['reporting'] = {'groups': new_reporting}

        elif instance['template']['name'] == 'factory':
            instance['outputs']['diff'] = instance['reporting']
            instance['reporting'] = None

        if instance['reporting'] and 'contacts' in instance['reporting']:
            if isinstance(instance['reporting']['contacts'], dict):
                instance['reporting']['contacts'] = list(instance['reporting']['contacts'].values())

        return instance

    @timed
    async def _migrate_task_instances(self):
        """
        Replace documents with new structure.
        """
        old_col = self.db['task-instances']
        new_col = self.db['task_instances']
        migrator = Migrator(old_col, new_col)
        async for task in migrator:
            # Convert task instance.
            instance = self._new_task(task)
            # Insert.
            migrator.new.insert(instance)
            migrator.old.find({'_id': instance['_id']}).remove_one()
        log.info('%s task instances migrated', migrator.count)

    @timed
    async def _migrate_old_instances(self):
        """
        Bring back the old 'instances' collection from the dead.
        These documents are way more older than the others, so the migrations
        are a bit more complex and specific.
        """
        task_count = 0
        old_col = self.db['instances']
        workflow_col = self.db['workflow_instances']
        task_col = self.db['task_instances']
        migrator = Migrator(old_col, workflow_col)
        async for workflow in migrator:
            # Split tasks from their workflow instance.
            tasks = workflow.pop('tasks')
            # Convert workflow instance.
            instance = workflow.pop('exec')
            instance['_id'] = workflow.pop('_id')
            instance['template'] = workflow
            task_count += len(tasks)

            # Convert and insert task instances.
            for index, task in enumerate(tasks):
                if task['exec']:
                    # Set inputs and outputs as dicts (could be None).
                    for io in ('inputs', 'outputs'):
                        if not isinstance(task['exec'][io], dict):
                            task['exec'][io] = {}

                # Migrate from the old send_email task reporting/config.
                if task['name'] == 'send_email':
                    config = task['config']
                    task['config'] = {
                        'recipients': {'source': 'field', 'value': 'recipients'},
                        'subject': config['subject'],
                        'message': config['message'],
                        'default_placeholder': config.get('default_placeholder', ''),
                        'sender': {'address': config.get('sender')},
                    }

                    if task['exec'] and task['exec']['reporting'] is not None:
                        old_reporting = task['exec']['reporting']
                        contacts_len = len(old_reporting['contacts'])
                        reporting = {
                            'subject': old_reporting['subject'],
                            'message': old_reporting['message'],
                            'contacts': [],
                            'delivery': {
                                'expected': contacts_len,
                                'sent': 0,
                                'error': 0,
                            },
                        }
                        # Contacts
                        for contact in old_reporting['contacts'].values():
                            new_contact = {
                                'display_name': ' '.join([contact.pop('last_name', ''), contact.pop('first_name', '')]),
                                'external': False,
                                'email': contact.get('email'),
                            }
                            if contact['state'] == 'sent':
                                reporting['delivery']['sent'] += 1
                                new_contact['state'] = 'sent'
                            elif contact['state'] in ('pending', 'error'):
                                new_contact['state'] = 'error'
                                reporting['delivery']['error'] += 1
                            reporting['contacts'].append(new_contact)

                        task['exec']['reporting'] = reporting

                # Migrate from the old send_sms task reporting/config.
                elif task['name'] == 'send_sms':
                    config = task['config']
                    task['config'] = {
                        'recipients': {'source': 'field', 'value': 'recipients'},
                        'message': config['message'],
                        'encoding': config.get('encoding', 'utf-8'),
                        'max_sms': config.get('max_sms'),
                        'default_placeholder': config.get('default_placeholder', ''),
                    }

                    if task['exec'] and task['exec']['reporting'] is not None:
                        old_reporting = task['exec']['reporting']
                        contacts_len = len(old_reporting['contacts'])
                        reporting = {
                            'message': old_reporting['message'],
                            'contacts': [],
                            'delivery': {
                                'expected': contacts_len,
                                'sent': 0,
                                'error': 0,
                            },
                        }
                        # Contacts
                        for contact in old_reporting['contacts'].values():
                            new_contact = {
                                'display_name': ' '.join([contact.pop('last_name', ''), contact.pop('first_name', '')]),
                                'external': False,
                                'mobile': contact.get('mobile'),
                            }
                            if contact['state'] == 'started':
                                reporting['delivery']['sent'] += 1
                                new_contact['state'] = 'sent'
                            elif contact['state'] == 'done':
                                new_contact['state'] = 'delivered'
                            elif contact['state'] == 'error':
                                reporting['delivery']['error'] += 1
                                new_contact['state'] = 'error'
                            reporting['contacts'].append(new_contact)

                        task['exec']['reporting'] = reporting

                # Migrate from the old wait_input task reporting/config.
                elif task['name'] in ('wait_sms', 'wait_email', 'wait_call'):
                    config = task['config']
                    task['config'] = {
                        'emitters': {'source': 'field', 'value': 'emitters'},
                        'rules': config['rules'],
                    }
                    if 'timeout' in config:
                        # This field will go away shortly after.
                        task['config']['blocking'] = {'timeout': config['timeout']}
                    # 'success_condition' => 'quorum_policy'
                    if 'success_condition' in config:
                        task['config']['quorum_policy'] = {
                            'method': 'count' if config['success_condition']['type'] == 'number' else 'rate',
                            'value': config['success_condition']['value'],
                        }

                    if task['exec'] and task['exec']['reporting'] is not None and 'emitters' in task['exec']['reporting']:
                        old_reporting = task['exec']['reporting']
                        contacts_len = len(old_reporting['emitters'])
                        reporting = {
                            'contacts': [],
                            'inputs': {
                                'expected': contacts_len,
                                'progress': 0,
                                'received': {
                                    'positive': 0,
                                    'negative': 0,
                                    'unknown': 0,
                                },
                                'failed': 0
                            }
                        }
                        # Quorum
                        quorum_policy = task['config'].get('quorum_policy')
                        if quorum_policy:
                            reporting['quorum'] = {
                                'reached': False,
                                'expected': ceil(quorum_policy['value'] if quorum_policy['method'] == 'count' else (contacts_len * quorum_policy['value']) / 100),
                                'progress': len(old_reporting['inputs_accepted']),
                            }
                            reached = (reporting['quorum']['progress'] >= reporting['quorum']['expected'])
                            reporting['quorum']['reached'] = reached
                            task['exec']['outputs']['quorum'] = reached

                        # Contacts
                        positives = {contact['source']: contact['time'] for contact in old_reporting['inputs_accepted']}
                        unknowns = {contact['source']: contact['time'] for contact in old_reporting['inputs_invalid']}
                        for contact in old_reporting['emitters']:
                            new_contact = {
                                'uid': contact['uid'],
                                'external': False,
                                'display_name': ' '.join([contact.get('last_name', ''), contact.get('first_name', '')]),
                                'input': None,
                                'source': None,
                                'received_at': None,
                                'error': False,
                            }

                            result = None
                            if task['name'] == 'wait_sms':
                                source = contact.get('mobile')
                                if source in positives.keys():
                                    result = 'positive'
                                    received_at = positives[source]
                                elif source in unknowns:
                                    result = 'unknown'
                                    received_at = unknowns[source]
                            elif task['name'] == 'wait_email':
                                source = contact.get('email')
                                if source in positives:
                                    result = 'positive'
                                    received_at = positives[source]
                                elif source in unknowns:
                                    result = 'unknown'
                                    received_at = unknowns[source]
                            elif task['name'] == 'wait_call':
                                for media in ('mobile', 'phone_work', 'phone_home'):
                                    source = contact.get(media)
                                    if source in positives:
                                        result = 'positive'
                                        received_at = positives[source]
                                    elif source in unknowns:
                                        result = 'unknown'
                                        received_at = unknowns[source]
                                    break

                            if result is None:
                                result = 'unknown'
                                received_at = task['exec']['end']

                            new_contact['input'] = result
                            new_contact['received_at'] = received_at
                            reporting['inputs']['progress'] += 1
                            reporting['inputs']['received'][result] += 1
                            reporting['contacts'].append(new_contact)

                        task['exec']['reporting'] = reporting

                # Migrate from the old call task reporting/config.
                elif task['name'] == 'call':
                    config = task['config']
                    task['config'] = {
                        'recipients': {'source': 'field', 'value': 'recipients'},
                        'template': {'id': config['template'], 'draft': config.get('draft', False)},
                    }
                    if 'global_timeout' in config:
                        # This field will go away shortly after.
                        task['config']['blocking'] = {'timeout': config['global_timeout']}
                    # 'success_condition' => 'quorum_policy'
                    if 'success_condition' in config:
                        task['config']['quorum_policy'] = {
                            'method': 'count' if config['success_condition']['type'] == 'number' else 'rate',
                            'value': config['success_condition']['value'],
                        }
                    if task['exec'] and task['exec']['reporting'] is not None:
                        old_reporting = task['exec']['reporting']
                        contacts_len = len(old_reporting['contacts'])
                        reporting = {
                            'contacts': [],
                            'feedbacks': {
                                'expected': contacts_len,
                                'progress': contacts_len,
                                'received': {
                                    'positive': 0,
                                    'negative': 0,
                                    'unknown': 0,
                                },
                                'failed': 0
                            }
                        }
                        # Quorum
                        quorum_policy = task['config'].get('quorum_policy')
                        if quorum_policy:
                            reporting['quorum'] = {
                                'reached': False,
                                'expected': ceil(quorum_policy['value'] if quorum_policy['method'] == 'count' else (contacts_len * quorum_policy['value']) / 100),
                                'progress': old_reporting['promises']['count'],
                            }
                            reached = (reporting['quorum']['progress'] >= reporting['quorum']['expected'])
                            reporting['quorum']['reached'] = reached
                            task['exec']['outputs']['quorum'] = reached
                        # Contacts
                        for contact in old_reporting['contacts'].values():
                            contact['display_name'] = ' '.join([contact.pop('last_name', ''), contact.pop('first_name', '')])
                            contact['external'] = False
                            contact['attempts'] = 1
                            contact['feedback'] = 'unknown'
                            contact['calls'] = []
                            medias = contact.pop('medias', {})
                            for media, call in medias.items():
                                if call.get('promise') is True:
                                    contact['feedback'] = 'positive'
                                contact['calls'].append({
                                    'uid': str(uuid4())[:8],
                                    'media': media,
                                    'number': contact.pop(media, None),
                                    'attempt': 1,
                                    'feedback': 'positive' if call.get('promise') is True else 'unknown',
                                    'status': 'completed',
                                    'workflow': call.get('workflow_exec_id'),
                                    'start': call['start'],
                                    'end': call['end'] if 'end' in call else None,
                                })
                            contact.pop('email', None)
                            reporting['contacts'].append(contact)
                            reporting['feedbacks']['received'][contact['feedback']] += 1

                        task['exec']['reporting'] = reporting

                tasks[index] = self._new_task(task, instance['id'])

            try:
                await task_col.insert_many(tasks)
            except BulkWriteError:
                pass

            # Insert workflow instance.
            migrator.new.insert(instance)
            migrator.old.find({'_id': instance['_id']}).remove_one()

        log.info(
            '%s old instances migrated to new format (including %s tasks)',
            migrator.count, task_count,
        )


if __name__ == '__main__':
    logging.basicConfig(format='%(message)s', level='DEBUG')
    m = Migration('localhost', 'twilio')
    asyncio.get_event_loop().run_until_complete(m.run())



import asyncio

from nyuki.utils import from_isoformat

from .api import Response, resource


@resource('/bus/topics', versions=['v1'])
class ApiBusTopics:

    async def get(self, request):
        try:
            self.nyuki._services.get('bus')
        except KeyError:
            return Response(status=404)
        return Response(self.nyuki.bus.topics)


@resource('/bus/publish', versions=['v1'])
class ApiBusPublish:

    async def post(self, request):
        try:
            self.nyuki._services.get('bus')
        except KeyError:
            return Response(status=404)
        request = await request.json()
        asyncio.ensure_future(self.nyuki.bus.publish(
            request.get('data', {}),
            request.get('topic', 'testing'),
            request.get('qos', 0),
        ))



import json
import asyncio
import logging
import pickle
import aiohttp
from uuid import uuid4
from copy import deepcopy
from random import shuffle
from datetime import datetime
from tukio import Engine, TaskRegistry, get_broker, EXEC_TOPIC
from tukio.workflow import Workflow, WorkflowExecState
from tukio.task.factory import TaskExecState

from nyuki import Nyuki
from nyuki.memory import memsafe
from nyuki.utils import serialize_object, utcnow
from nyuki.workflow.db.storage import MongoStorage
from nyuki.workflow.db.migrations import run_migrations
from nyuki.workflow.db.task_instances import WS_FILTERS

from .api.factory import (
    ApiFactoryRegex, ApiFactoryRegexes, ApiFactoryLookup, ApiFactoryLookups,
    ApiFactoryLookupCSV
)
from .api.templates import (
    ApiTasks, ApiTemplates, ApiTemplate, ApiTemplateVersion, ApiTemplateDraft
)
from .api.instances import (
    ApiWorkflow, ApiWorkflows, ApiWorkflowsHistory, ApiWorkflowHistory,
    ApiWorkflowTriggers, ApiWorkflowTrigger, ApiWorkflowHistoryTask,
    ApiWorkflowHistoryTaskData, ApiTaskReporting, ApiTaskReportingContact,
    ApiTaskReportingContacts
)
from .api.vars import (
    ApiVars, ApiVarsVersion, ApiVarsDraft
)

from .tasks import *
from .tasks.utils import runtime, CONTACT_PROGRESS
from .tukio import WorkflowSelector


log = logging.getLogger(__name__)


class BadRequestError(Exception):
    pass


@serialize_object.register(Workflow)
def _serialize_workflow(wf):
    """
    Workflow serializer.
    """
    return wf.report()


def sanitize_workflow_exec(obj):
    """
    Replace any object value by 'internal data' string to store in Mongo.
    """
    types = [dict, list, tuple, str, int, float, bool, type(None), datetime]
    if type(obj) not in types:
        obj = 'Internal server data: {}'.format(type(obj))
    elif isinstance(obj, dict):
        for key, value in obj.items():
            obj[key] = sanitize_workflow_exec(value)
    elif isinstance(obj, list):
        for item in obj:
            item = sanitize_workflow_exec(item)
    return obj


class WorkflowInstance:

    """
    Holds a workflow pair of template/instance.
    Allows retrieving a workflow exec state at any moment.
    """

    __slots__ = ('_template', '_instance', '_exec')

    ALLOWED_EXEC_KEYS = ['requester', 'track']

    def __init__(self, template, instance, **kwargs):
        self._template = template
        self._instance = instance
        self._exec = {
            key: kwargs[key]
            for key in kwargs
            if key in self.ALLOWED_EXEC_KEYS
        }

    @property
    def template(self):
        return self._template

    @property
    def instance(self):
        return self._instance

    @property
    def exec(self):
        return self._exec

    def report(self, tasks=True, data=True):
        """
        Merge a workflow exec instance report and its template.
        """
        template = deepcopy(self._template)
        inst = self._instance.report()
        inst['exec'].update(self._exec)

        result = {
            **inst['exec'],
            'template': template,
        }

        if tasks is False:
            del result['template']['graph']
            del result['template']['tasks']
            return result

        tasks = {task['id']: {'template': task} for task in template['tasks']}
        for task_dict in inst['tasks']:
            if not task_dict.get('exec'):
                # Task was never started, create dummy exec dict.
                task_dict['exec'] = {
                    'id': str(uuid4()),
                    'start': None,
                    'end': None,
                    'state': 'not-started',
                    'inputs': None,
                    'outputs': None,
                    'reporting': None
                }

            # Filter out reporting/data if not necessary
            if data is False:
                del task_dict['exec']['reporting']
                del task_dict['exec']['inputs']
                # Leave the necessary task-end informations available
                if task_dict['exec']['outputs']:
                    task_dict['exec']['outputs'] = {
                        key: task_dict['exec']['outputs'][key]
                        for key in WS_FILTERS
                        if key in task_dict['exec']['outputs']
                    }

            # Add execution informations to each task.
            tasks[task_dict['id']].update(task_dict['exec'])

        result['template']['tasks'] = list(tasks.values())
        return result


class WorkflowNyuki(Nyuki):

    """
    Generic workflow nyuki allowing data storage and manipulation
    of tukio's workflows.
    https://github.com/optiflows/tukio
    """

    CONF_SCHEMA = {
        'type': 'object',
        'required': ['mongo'],
        'properties': {
            'mongo': {
                'type': 'object',
                'required': ['host', 'database'],
                'properties': {
                    'host': {'type': 'string', 'minLength': 1},
                    'database': {'type': 'string', 'minLength': 1},
                    'validate_on_start': {'type': 'boolean', 'default': True},
                }
            },
            'topics': {
                'type': 'array',
                'items': {'type': 'string', 'minLength': 1}
            }
        }
    }
    HTTP_RESOURCES = Nyuki.HTTP_RESOURCES + [
        ApiTasks,                   # /v1/workflow/tasks
        ApiTemplates,               # /v1/workflow/templates
        ApiTemplate,                # /v1/workflow/templates/{uid}
        ApiTemplateDraft,           # /v1/workflow/templates/{uid}/draft
        ApiTemplateVersion,         # /v1/workflow/templates/{uid}/{version}
        ApiWorkflows,               # /v1/workflow/instances
        ApiWorkflow,                # /v1/workflow/instances/{uid}
        ApiTaskReporting,           # /v1/workflow/instances/{uid}/tasks/{task_id}/reporting
        ApiTaskReportingContacts,   # /v1/workflow/instances/{uid}/tasks/{task_id}/reporting/contacts
        ApiTaskReportingContact,    # /v1/workflow/instances/{uid}/tasks/{task_id}/reporting/contacts/{contact_id}
        ApiWorkflowsHistory,        # /v1/workflow/history
        ApiWorkflowHistory,         # /v1/workflow/history/{uid}
        ApiWorkflowHistoryTask,     # /v1/workflow/history/{uid}/tasks/{task_id}
        ApiWorkflowHistoryTaskData, # /v1/workflow/history/{uid}/tasks/{task_id}/data
        ApiFactoryRegexes,          # /v1/workflow/regexes
        ApiFactoryRegex,            # /v1/workflow/regexes/{uid}
        ApiFactoryLookups,          # /v1/workflow/lookups
        ApiFactoryLookup,           # /v1/workflow/lookups/{uid}
        ApiFactoryLookupCSV,        # /v1/workflow/lookups/{uid}/csv
        ApiWorkflowTriggers,        # /v1/workflow/triggers
        ApiWorkflowTrigger,         # /v1/workflow/triggers/{tid},
        ApiVars,                    # /v1/workflow/vars/{uid}
        ApiVarsVersion,             # /v1/workflow/vars/{uid}/{version}
        ApiVarsDraft,               # /v1/workflow/data/{uid}/draft
    ]

    DEFAULT_POLICY = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.migrate_config()
        self.register_schema(self.CONF_SCHEMA)
        self.engine = None
        self.storage = MongoStorage()

        self.AVAILABLE_TASKS = {}
        for name, value in TaskRegistry.all().items():
            self.AVAILABLE_TASKS[name] = getattr(value[0], 'SCHEMA', {})

        # Stores workflow instances with their template data
        self.running_workflows = {}

        runtime.bus = self.bus
        runtime.config = self.config
        runtime.workflows = self.running_workflows

    @property
    def mongo_config(self):
        return self.config['mongo']

    @property
    def topics(self):
        return self.config.get('topics', [])

    async def setup(self):
        self.storage.configure(**self.mongo_config)
        # Blocks until connection to Mongo is done.
        await self.storage.index()
        await run_migrations(**self.mongo_config)
        selector = WorkflowSelector(self.storage)
        self.engine = Engine(selector=selector, loop=self.loop)
        for topic in self.topics:
            asyncio.ensure_future(self.bus.subscribe(
                topic, self.workflow_event
            ))
        # Enable workflow exec follow-up
        get_broker().register(self.report_workflow, topic=EXEC_TOPIC)
        # Handle distributed workflow's failures
        if 'raft' in self._services.all:
            self.raft.register('failures', self.failure_handler)

    async def reload(self):
        self.storage.configure(**self.mongo_config)

    async def teardown(self):
        if self.engine:
            await self.engine.stop()

    def new_workflow(self, template, instance, **kwargs):
        """
        Keep in memory a workflow template/instance pair.
        """
        wflow = WorkflowInstance(template, instance, **kwargs)
        self.running_workflows[instance.uid] = wflow
        if 'memory' in self._services and self.memory.available:
            asyncio.ensure_future(
                self.write_report(wflow.report(), False)
            )
        return wflow

    async def report_workflow(self, event):
        """
        Send all worklfow updates to the clients.
        """
        source = event.source.as_dict()
        instance_id = source['workflow_exec_id']
        try:
            wflow = self.running_workflows[instance_id]
        except KeyError:
            log.debug('Outdated event to report: %s', event)
            return

        topic = 'workflow/exec/{}'.format(instance_id)
        payload = {
            'type': event.data['type'],
            'ts': utcnow(),
            'topic': topic,
            'service': self.bus.name,
        }

        # If is a workflow update, send the 'child' value
        if event.data['type'] in WorkflowExecState.values():
            requester = wflow.exec.get('requester')
            if requester:
                payload['requester'] = requester

        # A task information requires the corresponding template id
        # and a more precise topic.
        task_exec_id = source.get('task_exec_id')
        if task_exec_id:
            topic = '{}/tasks/{}'.format(topic, task_exec_id)
            payload['template'] = {'id': source['task_template_id']}

            # Append full data if this is a 'task-progress'
            if event.data['type'] == TaskExecState.PROGRESS.value:
                payload['data'] = event.data.get('content') or {}
                topic = '{}/reporting'.format(topic)

            # Custom event type for single contact updates
            elif event.data['type'] == CONTACT_PROGRESS:
                payload['data'] = event.data.get('content') or {}
                topic = '{}/reporting/contacts/{}'.format(
                    topic, payload['data']['uid'],
                )
                # Remove contact uid (available in topic)
                del payload['data']['uid']

            elif event.data['type'] in (
                TaskExecState.TIMEOUT.value,
                TaskExecState.END.value,
                TaskExecState.ERROR.value,
            ):
                if isinstance(event.data['content'], dict):
                    # Only send the task's important fields, if any
                    payload['data'] = {
                        key: event.data['content'][key]
                        for key in WS_FILTERS
                        if key in event.data['content']
                    }

            # Update topic for this event
            payload['topic'] = topic

        memwrite = True
        # Workflow begins, also send the full template.
        if event.data['type'] == WorkflowExecState.BEGIN.value:
            payload['template'] = dict(wflow.template)
        # Workflow ended, clear it from memory
        elif event.data['type'] in [
            WorkflowExecState.END.value,
            WorkflowExecState.ERROR.value
        ]:
            payload['data'] = event.data.get('content') or {}
            # Sanitize objects to store the finished workflow instance
            asyncio.ensure_future(self.storage.insert_instance(
                sanitize_workflow_exec(wflow.report())
            ))
            del self.running_workflows[instance_id]
            memwrite = False

        # Shared memory set/del
        if 'memory' in self._services and self.memory.available:
            if memwrite:
                memjob = self.write_report(wflow.report())
            else:
                memjob = self.clear_report(instance_id)
            asyncio.ensure_future(memjob)

        await self.bus.publish(payload, 'websocket/{}'.format(topic))

    async def workflow_event(self, topic, data):
        """
        New bus event received, trigger workflows if needed.
        """
        templates = {}
        # Retrieve full workflow templates
        # TODO: Better way to fetch the full template details
        # (this does may more requests than it should)
        wf_templates = await self.engine.selector.select(topic)
        for wftmpl in wf_templates:
            templates[wftmpl.uid] = await self.storage.get_template(
                wftmpl.uid, draft=False
            )
        # Trigger workflows
        instances = await self.engine.data_received(data, topic)
        for instance in instances:
            self.new_workflow(templates[instance.template.uid], instance)

    @memsafe
    async def failure_handler(self, instances):
        """
        This method is called upon failure detection implemented in raft.

        Failing instances are just instances that have been shutdown, either
        because of a crash, or a restart. Let's find out which ones need a
        failover (basically, rescuing workflows detached from dead instances).
        """
        log.debug(
            "The following '%s' instances are failing: %s",
            self.config['service'], instances
        )

        # Select eligible rescuers
        ntw = self.raft.network
        rescuers = [ipv4 for ipv4, uid in ntw.items() if uid not in instances]

        # Iterate over all failing instances
        for ifrom in instances:

            # Fetch the list of workflows for a given failing instance.
            index = self.memory.key(ifrom, 'workflows', 'instances')
            for wflow in await self.memory.store.smembers(index):
                wflow = wflow.decode('utf-8')

                # Get the report shared by the failing instance
                try:
                    report = await self.read_report(wflow, ifrom)
                except KeyError:
                    log.error("Workflow %s memory has been wiped out", wflow)
                    break

                shuffle(rescuers)
                report = json.dumps(report, default=serialize_object)

                # Send a failover request to a valid, not failing, instance.
                for ito in rescuers:
                    request = {
                        'url': 'http://{}:{}/v1/workflow/instances'.format(
                            ito, self.api._port
                        ),
                        'headers': {'Content-Type': 'application/json'},
                        'data': report
                    }
                    async with aiohttp.ClientSession() as session:
                        async with session.put(**request) as resp:
                            if resp.status == 200:
                                # `ito` rescuer has taken over the workflow
                                break
                else:
                    log.error("Workflow %s hasn't be rescued properly", wflow)
                    continue
                asyncio.ensure_future(self.clear_report(wflow, ifrom=ifrom))

    @memsafe
    async def clear_report(self, uid, ifrom=None):
        """
        Remove a report from the shared memory.
        """
        _iform = ifrom or self.id
        await self.memory.store.delete(
            key=self.memory.key(_iform, 'workflows', 'instances', uid)
        )
        await self.memory.store.srem(
            key=self.memory.key(_iform, 'workflows', 'instances'),
            member=uid
        )

    @memsafe
    async def write_report(self, report, replace=True, ito=None):
        """
        Store an instance report into shared memory.
        A simple 'set' is used againts a 'hset' (hash storage), even though the
        'hset' seems more appropriate, because a field in a hash can't have TTL
        """
        _ito = ito or self.id
        uid = report['exec']['id']
        response = await self.memory.store.set(
            key=self.memory.key(_ito, 'workflows', 'instances', uid),
            value=pickle.dumps(report),
            expire=86400,
            exist=None if replace else False
        )

        if not response:
            log.error("Can't share workflow id %s context in memory", uid)
            return

        keyspace = self.memory.key(self.id, 'workflows', 'instances')
        await self.memory.store.sadd(key=keyspace, member=uid)
        await self.memory.store.expire(key=keyspace, timeout=86400)

    @memsafe
    async def read_report(self, uid, ifrom=None):
        """
        Read and parse a report from the shared memory.
        """
        _iform = ifrom or self.id
        report = await self.memory.store.get(
            key=self.memory.key(_iform, 'workflows', 'instances', uid)
        )
        if not report:
            raise KeyError("Can't find workflow id context %s in memory", uid)
        return pickle.loads(report)
